@@ -17,7 +17,7 @@ Util::H2O - Hash to Object: turns hashrefs into objects with accessors for keys
  my $hash = h2o { foo => "bar", x => "y" }, qw/ more keys /;
  print $hash->foo, "\n";           # accessor
  $hash->x("z");                    # change value
- $hash->more("quz");               # additional keys
+ $hash->more("cowbell");           # additional keys
  
  my $struct = { hello => { perl => "world!" } };
  h2o -recurse, $struct;            # objectify nested hashrefs as well
@@ -40,7 +40,7 @@ Util::H2O - Hash to Object: turns hashrefs into objects with accessors for keys
 
 =cut
 
-our $VERSION = '0.10';
+our $VERSION = '0.12';
 # For AUTHOR, COPYRIGHT, AND LICENSE see the bottom of this file
 
 our @EXPORT = qw/ h2o /;  ## no critic (ProhibitAutomaticExportation)
@@ -52,8 +52,12 @@ BEGIN {
 	# uncoverable condition false
 	if ( $] ge '5.008009' ) {
 		require Hash::Util;
-		Hash::Util->import('lock_ref_keys') }
-	else { *lock_ref_keys = sub {} }  # uncoverable statement
+		Hash::Util->import(qw/ lock_ref_keys lock_hashref /) }
+	else {
+		*lock_ref_keys = *lock_hashref = sub {
+			carp "this Perl is too old to lock the hash";  # uncoverable statement
+		};  # uncoverable statement
+	}
 }
 
 =head1 Description
@@ -81,9 +85,14 @@ one will take effect.
 
 =item C<-recurse>
 
-Nested hashes are objectified as well. Note that I<none> of the other
-options will be applied to the nested hashes, including
-C<@additional_keys>. Nested arrayrefs are not recursed into.
+Nested hashes are objectified as well. The only options that are passed down to
+nested hashes are C<-lock> and C<-ro>. I<None> of the other options will be
+applied to the nested hashes, including C<@additional_keys>. Nested arrayrefs
+are not recursed into.
+
+Versions of this module before v0.12 did not pass down the C<-lock> option,
+meaning that if you used C<-nolock, -recurse> on those versions, the nested
+hashes would still be locked.
 
 =item C<-meth>
 
@@ -136,18 +145,36 @@ in Perl's symbol table and use memory accordingly.
 
 Whether or not to use L<Hash::Util>'s C<lock_ref_keys> to prevent
 modifications to the hash's keyset. Defaults to I<true>.
+The C<-nolock> option is provided as a short form of C<< -lock=>0 >>.
 
 Keysets of objects created by the constructor generated by the
-C<-new> option are I<not> locked.
+C<-new> option are also locked. Versions of this module before
+v0.12 did not lock the keysets of new objects.
 
 Note that on really old Perls, that is, before Perl v5.8.9,
 L<Hash::Util> and its C<lock_ref_keys> are not available, so the hash
 is never locked on those versions of Perl. Versions of this module
 before v0.06 did not lock the keyset.
+Versions of this module as of v0.12 issue a warning on old Perls.
 
 =item C<-nolock>
 
-Short form of the option C<< lock=>0 >>.
+Short form of the option C<< -lock=>0 >>.
+
+=item C<-ro>
+
+Makes the entire hash read-only using L<Hash::Util>'s C<lock_hashref> and the
+generated accessors will also throw an error if you try to change values. In
+other words, this makes the object and the underlying hash immutable.
+
+You cannot specify any C<@additional_keys> with this option enabled unless you
+also use the C<-new> option - the additional keys will then only be useful as
+arguments to the constructor. This option can't be used with C<-nolock> or
+C<< -lock=>0 >>.
+
+This option was added in v0.12. Using this option will not work and cause a
+warning when used on really old Perls (before v5.8.9), because this
+functionality was not yet available there.
 
 =back
 
@@ -156,7 +183,8 @@ Short form of the option C<< lock=>0 >>.
 You must supply a plain (unblessed) hash reference here. Be aware
 that this function I<does> modify the original hashref(s) by blessing
 it and locking its keyset (the latter can be disabled with the
-C<-lock> option).
+C<-lock> option), and if you use C<-meth> or C<-classify>, keys whose
+values are code references will be removed.
 
 An accessor will be set up for each key in the hash; note that the
 keys must of course be valid Perl identifiers for you to be able to
@@ -165,6 +193,7 @@ call the method normally.
 When C<-clean> is I<true> (the default, unless you use C<-class>),
 the hash may not contain a key named C<DESTROY>. When C<-new> is
 used, the hash may not contain a key named C<new>.
+If the hash contains a key named C<AUTOLOAD>, see L</AUTOLOAD>.
 
 =head3 C<@additional_keys>
 
@@ -177,13 +206,14 @@ The (now blessed and optionally locked) C<$hashref>.
 =cut
 
 sub h2o {  ## no critic (RequireArgUnpacking, ProhibitExcessComplexity)
-	my ($recurse,$meth,$class,$new,$clean,$lock);
+	my ($recurse,$meth,$class,$new,$clean,$lock,$ro);
 	while ( @_ && $_[0] && !ref$_[0] ) {
 		if ($_[0] eq '-recurse' ) { $recurse = shift }  ## no critic (ProhibitCascadingIfElse)
 		elsif ($_[0] eq '-meth' ) { $meth    = shift }
 		elsif ($_[0] eq '-clean') { $clean   = (shift, shift()?1:0) }
 		elsif ($_[0] eq '-lock' ) { $lock    = (shift, shift()?1:0) }
 		elsif ($_[0] eq '-nolock'){ $lock = 0; shift }
+		elsif ($_[0] eq '-ro'   ) { $ro      = shift }
 		elsif ($_[0] eq '-new'  ) { $new     = shift }
 		elsif ($_[0] eq '-class') {
 			$class = (shift, shift);
@@ -202,16 +232,20 @@ sub h2o {  ## no critic (RequireArgUnpacking, ProhibitExcessComplexity)
 	$lock = 1 unless defined $lock;
 	my $hash = shift;
 	croak "h2o must be given a plain hashref" unless ref $hash eq 'HASH';
+	croak "h2o with additional keys doesn't make sense with -ro" if $ro && @_ && !$new;
 	my %ak   = map {$_=>1} @_;
 	my %keys = map {$_=>1} @_, keys %$hash;
 	croak "h2o hashref may not contain a key named DESTROY"
 		if $clean && exists $keys{DESTROY};
 	croak "h2o hashref may not contain a key named new if you use the -new option"
 		if $new && exists $keys{new};
-	if ($recurse) { ref eq 'HASH' and h2o(-recurse,$_) for values %$hash }
+	croak "h2o can't turn off -lock if -ro is on" if $ro && !$lock;
+	if ($recurse) { ref eq 'HASH' and h2o(-recurse,-lock=>$lock,($ro?-ro:()),$_) for values %$hash }
 	my $pack = defined $class ? $class : sprintf('Util::H2O::_%x', $hash+0);
 	for my $k (keys %keys) {
-		my $sub = sub { my $self = shift; $self->{$k} = shift if @_; $self->{$k} };
+		my $sub = $ro
+			? sub { my $self = shift; croak "this object is read-only" if @_; exists $self->{$k} ? $self->{$k} : undef }
+			: sub { my $self = shift; $self->{$k} = shift if @_; $self->{$k} };
 		if ( $meth && ref $$hash{$k} eq 'CODE' )
 			{ $sub = delete $$hash{$k}; $ak{$k} or delete $keys{$k} }
 		{ no strict 'refs'; *{"${pack}::$k"} = $sub }  ## no critic (ProhibitNoStrict)
@@ -225,20 +259,38 @@ sub h2o {  ## no critic (RequireArgUnpacking, ProhibitExcessComplexity)
 			my $class = shift;
 			$class = ref $class if ref $class;
 			croak "Odd number of elements in argument list" if @_%2;
-			my %self = @_;
-			exists $keys{$_} or croak "Unknown argument '$_'" for keys %self;
-			return bless \%self, $class;
+			my $self = {@_};
+			exists $keys{$_} or croak "Unknown argument '$_'" for keys %$self;
+			bless $self, $class;
+			if ($ro) { lock_hashref $self }
+			elsif ($lock) { lock_ref_keys $self, keys %keys }
+			return $self;
 		};
 		{ no strict 'refs'; *{$pack.'::new'} = $sub }  ## no critic (ProhibitNoStrict)
 	}
 	bless $hash, $pack;
-	lock_ref_keys $hash, keys %keys if $lock;
+	if ($ro) { lock_hashref $hash }
+	elsif ($lock) { lock_ref_keys $hash, keys %keys }
 	return $hash;
 }
 
 1;
 __END__
 
+=head1 Notes
+
+=head2 C<AUTOLOAD>
+
+If your hash contains a key named C<AUTOLOAD>, or this key is present in
+C<@additional_keys>, this module will set up a method called C<AUTOLOAD>, which
+is subject to Perl's normal autoloading behavior - see L<perlsub/Autoloading>
+and L<perlobj/AUTOLOAD>. Without the C<-meth> option, you will get a
+"catch-all" accessor to which all method calls to unknown method names will go,
+and with C<-meth> enabled (which is implied by C<-classify>), you can install
+your own custom C<AUTOLOAD> handler by passing a coderef as the value for this
+key. However, it is important to note that enabling autoloading removes any
+typo protection on method names.
+
 =head1 See Also
 
 Inspired in part by C<lock_keys> from L<Hash::Util>.
@@ -256,7 +308,7 @@ For real OO work, I like L<Moo> and L<Type::Tiny>.
 
 =head1 Author, Copyright, and License
 
-Copyright (c) 2020 Hauke Daempfling (haukex@zero-g.net).
+Copyright (c) 2020-2021 Hauke Daempfling (haukex@zero-g.net).
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl 5 itself.
